#!BPY

"""
Name: 'NWN Model (.mdl)...'
Blender: 239
Group: 'Import'
Tooltip: 'Import an Aurora MDL file'
"""

__author__ = " Jean-Baptiste PERIN (jb_perin(at)yahoo.fr)"
__version__ = "1.2"
__url__ = ["Communicate problems and errors, http://www.zoo-logique.org/3D.Blender/newsportal/thread.php?group=3D.Blender"]
__email__= ["M. PERIN Jean-Baptiste, jb_perin@yahoo.fr", "scripts"]
__bpydoc__ = """\
mdl_import.py v1.2

Script loading an NWN/Aurora MDL File,  
Usage:<br>
 - Run the script <br>
 - Choose the file to open<br>
 - Press Neverwinter MDL file button<br>
this importer handles: <br>
- meshes <br>
- uv coordinates<br>
- textures <br>
- animations <br>
"""
#----------------------------------------------
# (c) Jean-Baptiste PERIN  december 2005, released under Blender Artistic Licence
#    for the Blender 2.40 Python Scripts Bundle.
#----------------------------------------------

import Blender
from Blender import *
import string
from string import *
import math
from Blender.sys import join, dirname, splitext, basename

NB_MAX_DATA_IN_LIST = 2000
ALTERNATE_IMAGE_DIRECTORY = 'C:\\Your\\Alternate\\Texture\\Directory\\Here'

dicObject = {}

#***************************
#**** MDL STRUCTURE ********
#***************************
def createRotationMatrix(anAxis, angle):
  axis = anAxis
  sine = math.sin(angle)
  cosine = math.cos(angle)
  one_minus_cosine=1-cosine
  axis.normalize ()
  matrice = Mathutils.Matrix([(one_minus_cosine * axis[0] * axis[0]) + cosine,(one_minus_cosine * axis[0] * axis[1]) - (axis[2] * sine),(one_minus_cosine * axis[2] * axis[0]) + (axis[1] * sine), 0],
                   [(one_minus_cosine * axis[0] * axis[1]) + (axis[2] * sine),(one_minus_cosine * axis[1] * axis[1]) + cosine,(one_minus_cosine * axis[1] * axis[2]) - (axis[0] * sine), 0],
           [(one_minus_cosine * axis[2] * axis[0]) - (axis[1] * sine),(one_minus_cosine * axis[1] * axis[2]) + (axis[0] * sine),(one_minus_cosine * axis[2] * axis[2]) + cosine, 0],
           [0,0 ,0 ,1.0])
  return matrice

def GetOrCreateCurve(ipo, curvename):
  """
  Retrieve or create a Blender Ipo Curve named C{curvename} in the C{ipo} Ipo

  >>> import mylib 

  >>> lIpo = GetOrCreateIPO("Une IPO")
  >>> laCurve = GetOrCreateCurve(lIpo, "RotX")

  Either an ipo curve named C{curvename} exists before the call then this curve is returned,
  Or such a curve doesn't exist before the call .. then it is created into the c{ipo} Ipo and returned 

  @type  ipo: Blender Ipo
  @param ipo: the Ipo in which the curve must be retrieved or created.
  @type  curvename: string
  @param curvename: name of the IPO.
  @rtype:   Blender Curve
  @return:  a Blender Curve named C{curvename} in the C{ipo} Ipo 
  """
  try:
    mycurve = ipo.getCurve(curvename)
    if mycurve != None:
      pass
    else:
      mycurve = ipo.addCurve(curvename)
  except:
    mycurve = ipo.addCurve(curvename)
  return mycurve
  
class Node:
  parent = 'NULL'
  name = "EmptyNode"
  type = 'Node'
  def __init__(self, nodename):
    self.name =  nodename
    self.orientation = [[0.0, 0.0, 0.0],0.0]
  def setParent(self,parentname):
    self.parent = parentname
  def setOrientation(self, rot):
    self.orientation = rot



class MeshNode(Node):
  def __init__(self, nodename):
    Node.__init__(self, nodename)
    self.position = [0.0, 0.0, 0.0]
    self.type = 'meshnode'

  def setPosition(self, pos):
    self.position = pos



class DummyNode(MeshNode):
  def __init__(self, nodename):
    MeshNode.__init__(self, nodename)
    self.type = 'dummynode'



class TrimeshNode(MeshNode):
  def __init__(self, nodename):
    MeshNode.__init__(self, nodename)
    self.bitmap = "noimage.tga"
    self.vertex = []
    self.uvcoords = []
    self.faces = []
    self.type = 'trimeshnode'
    self.alpha = 0

  def setBitmap (self, imagename):
    self.bitmap = imagename
    
  def setAlpha (self, val):
    self.alpha = val

  def addVertice(self, coords):
    self.vertex.append(coords)

  def addUVCoords(self, coords):
    self.uvcoords.append(coords)

  def addFace(self, vertind, uvind):
    self.faces.append([vertind, uvind])


class AnimNode(Node):
  def __init__(self, nodename):
    self.positionsKeys = []
    self.orientationsKeys = []
    self.type = 'animnode'
    Node.__init__(self, nodename)

  def addPositionsKey (self, time, poskey):
    self.positionsKeys.append([time, poskey])

  def addOrientationsKey (self, time, veckey, anglekey):
    self.orientationsKeys.append([time, veckey, anglekey])


class ModelGeom:
  def __init__(self,modelname):
    self.name = modelname
    self.nodes = {}

  def addNode(self, thenode):
    self.nodes [thenode.name] = thenode

class Anim:
  def __init__(self, animname, trimeshname):
    self.name = animname
    self.trimesh = trimeshname
    self.nodes = {}

  def addNode(self, thenode):
    self.nodes [thenode.name] = thenode


class MDLContent:
  name = 'EmptyContent'
  def __init__(self, modelname):
    self.name = modelname
    self.model = None
    self.anims = {}
  def setModel (self, geommodel):
    self.model = geommodel

  def addAnim(self, modelanim):
    self.anims[modelanim.name] = modelanim


#************************
#**** MDL PARSER ********
#************************

ST_START = 0

ST_READING_MESH = 1
ST_READING_ANIM = 2

ST_READING_VERTS = 3
ST_READING_TVERTS = 4
ST_READING_FACES = 5

ST_READING_ORIENTATION_KEYS = 6
ST_READING_POSITION_KEYS = 7

ST_READING_UNPROCESSED_NODE = 8


State = ST_START
OldState = ST_START
currentNode = None
currentModel = None
currentAnim = None
currentContent = None
nbDataToRead = 0


def ProcessLine(ligne):
  global State, OldState, currentNode, nbDataToRead, currentModel, currentContent, currentAnim
  champs = ligne.split()
  #print champs
  if len(champs)>= 1:

    if (champs[0] == 'newmodel') and (State != ST_READING_UNPROCESSED_NODE):
      currentContent = MDLContent(champs[1])

    elif (champs[0] == 'beginmodelgeom'):
      State = ST_READING_MESH
      #print "State = ST_READING_MESH"
      currentModel = ModelGeom(champs[1])

    elif (champs[0] == 'endmodelgeom') and (State != ST_READING_UNPROCESSED_NODE):
      currentContent.setModel(currentModel)

    elif (champs[0] == 'newanim') and (State != ST_READING_UNPROCESSED_NODE):
      State = ST_READING_ANIM
      #print "State = ST_READING_ANIM"
      #print 'newanim', champs[1], champs[2]
      currentAnim = Anim(champs[1], champs[2])

    elif (champs[0] == 'doneanim') and (State != ST_READING_UNPROCESSED_NODE):
      currentContent.addAnim(currentAnim)

    elif (champs[0] == 'node') and (State != ST_READING_UNPROCESSED_NODE):
      #print champs
      if State == ST_READING_MESH:
        if champs[1] == 'dummy':
          currentNode = DummyNode(champs[2])
        elif (champs[1] == 'trimesh')  or (champs[1] == 'danglymesh'):
      #print champs
          if len(champs) > 3: 
            currentNode = TrimeshNode(champs[2]+"_"+champs[3]) # what if len champs >4???
          else:
            currentNode = TrimeshNode(champs[2])
        elif (champs[1] == 'aabb') or (champs[1] == 'light') or (champs[1] == 'emitter'):
          #currentNode = TrimeshNode(champs[2])
          OldState = State
          State = ST_READING_UNPROCESSED_NODE
      elif State == ST_READING_ANIM:
        currentNode = AnimNode(champs[2])
      #print currentNode.name

    elif champs[0] == 'endnode':
      if State == ST_READING_MESH:
        currentModel.addNode(currentNode)
      elif State == ST_READING_ANIM:
        currentAnim.addNode(currentNode)
      elif State == ST_READING_UNPROCESSED_NODE:
        State = OldState

    elif (champs[0] == 'parent') and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.setParent(champs[1])

    elif (champs[0] == 'alpha') and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.setAlpha(1)

    elif champs[0] == 'bitmap' and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.setBitmap(champs[1])

    elif (champs[0] == 'position') and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.setPosition([float(champs[1]), float(champs[2]), float(champs[3])])
      #if currentNode.name == 'Box1291':
      #print "position de ", currentNode.name, " est ",[float(champs[1]), float(champs[2]), float(champs[3])] 
      #  print triNode.position

    elif (champs[0] == 'orientation') and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.setOrientation([[float(champs[1]), float(champs[2]), float(champs[3])], float(champs[4])])

    elif (champs[0] == 'verts') and (State != ST_READING_UNPROCESSED_NODE):
      State = ST_READING_VERTS
      nbDataToRead = int(champs[1])

    elif (champs[0] == 'tverts') and (State != ST_READING_UNPROCESSED_NODE):
      State = ST_READING_TVERTS
      nbDataToRead = int(champs[1])

    elif (champs[0] == 'faces')  and (State != ST_READING_UNPROCESSED_NODE):
      State = ST_READING_FACES
      nbDataToRead = int(champs[1])
    
    elif (champs[0] == 'colors')  and (State != ST_READING_UNPROCESSED_NODE):
      nbDataToRead = int(champs[1])
    elif (champs[0] == 'constraints')  and (State != ST_READING_UNPROCESSED_NODE):
      nbDataToRead = int(champs[1])

    elif (champs[0] == 'orientationkey')  and (State != ST_READING_UNPROCESSED_NODE):
      State = ST_READING_ORIENTATION_KEYS
      if len(champs) > 1:
        nbDataToRead = int(champs[1])
      else:
        nbDataToRead = NB_MAX_DATA_IN_LIST
    elif (champs[0] == 'positionkey')  and (State != ST_READING_UNPROCESSED_NODE):
      State = ST_READING_POSITION_KEYS
      if len(champs) > 1:
        nbDataToRead = int(champs[1])
      else:
        nbDataToRead = NB_MAX_DATA_IN_LIST
    elif (champs[0] == 'endlist') and (State != ST_READING_UNPROCESSED_NODE):
      nbDataToRead = 0
      if (State == ST_READING_POSITION_KEYS) or (State == ST_READING_ORIENTATION_KEYS):
        State = ST_READING_ANIM
      elif (State == ST_READING_VERTS) or (State == ST_READING_TVERTS) or (State == ST_READING_FACES):
        State = ST_READING_MESH

    elif (State == ST_READING_VERTS) and (nbDataToRead != 0) and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.addVertice([float(champs[0]), float(champs[1]), float(champs[2])])
      nbDataToRead -= 1
      if nbDataToRead == 0:
        State = ST_READING_MESH

    elif (State == ST_READING_TVERTS) and (nbDataToRead != 0)  and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.addUVCoords((float(champs[0]), float(champs[1])))
      nbDataToRead -= 1
      if nbDataToRead == 0:
        State = ST_READING_MESH

    elif (State == ST_READING_FACES) and (nbDataToRead != 0)  and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.addFace([int(champs[0]), int(champs[1]), int(champs[2])], [int(champs[4]), int(champs[5]), int(champs[6])])
      nbDataToRead -= 1
      if nbDataToRead == 0:
        State = ST_READING_MESH

    elif (State == ST_READING_ORIENTATION_KEYS) and (nbDataToRead != 0)  and (State != ST_READING_UNPROCESSED_NODE):
      currentNode.addOrientationsKey (float(champs[0]), [float(champs[1]), float(champs[2]), float(champs[3])], float(champs[4]))
      nbDataToRead -= 1
      if nbDataToRead == 0:
        State = ST_READING_ANIM

    elif (State == ST_READING_POSITION_KEYS) and (nbDataToRead != 0)  and (State != ST_READING_UNPROCESSED_NODE):
      #print champs
      currentNode.addPositionsKey (float(champs[0]), [float(champs[1]), float(champs[2]), float(champs[3])])
      nbDataToRead -= 1
      if nbDataToRead == 0:
        State = ST_READING_ANIM



filepath = ''
thefilename = ''  


#########
# Cette fonction renvoie un empty 
# in  : objname : le nom de l'empty recherche
# out : myobj : l'empty cree ou retrouve
#########
def CreateEmpty (objname):
  myobj = None
  myobj = Blender.Object.New('Empty',objname)
  Blender.Scene.getCurrent().link(myobj)
  return myobj

#########
# Cette fonction renvoie la liste des empties
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getAllEmpties():
    emp_list = []
    objs = Blender.Object.Get()
    for o in objs:
        if o.getType()=="Empty":
            emp_list.append(o)
    return emp_list
# 
# Creation d'un object image
# 
def GetOrCreateIMAGE (imagename, alpha):
  global filepath
  #if alpha == 1:
  #  filename = lower(imagename+'.tga')
  #else:
  #  filename=lower(imagename+'.jpg')
  filename = lower(imagename+'.tga')
  im=None

  try:
    im=Image.get(filename)
  except NameError:
    try:
      im=Image.Load(Blender.sys.join(filepath, filename))
    except IOError:
      try:
        im=Image.Load(Blender.sys.join(ALTERNATE_IMAGE_DIRECTORY, filename))
      except IOError:
        print "unable lo load image", filename
        pass
  
  return im



#***************************
#**** MESH CREATION ********
#***************************

def BuildTrimeshNode(triNode):
  me=NMesh.New(triNode.name)
  if triNode.bitmap != 'NULL':

    for ve in triNode.vertex:
      v=NMesh.Vert(ve[0],ve[1],ve[2])
      me.verts.append(v)
    me.name=triNode.name

    for fa in triNode.faces:
      f=NMesh.Face()  
      f.uv=[]
      f.v.append(me.verts[fa[0][0]])
      f.v.append(me.verts[fa[0][1]])
      f.v.append(me.verts[fa[0][2]])
      if (len(triNode.uvcoords)>0):
        f.uv.append(triNode.uvcoords[fa[1][0]])
        f.uv.append(triNode.uvcoords[fa[1][1]])
        f.uv.append(triNode.uvcoords[fa[1][2]])
      f.mode = NMesh.FaceModes.TEX | NMesh.FaceModes.LIGHT | NMesh.FaceModes.DYNAMIC 
      if triNode.alpha == 1:
        f.transp=NMesh.FaceTranspModes.ALPHA
      f.mat = 0
      f.smooth = 1
      if triNode.bitmap != 'NULL':
        im = GetOrCreateIMAGE(triNode.bitmap, triNode.alpha)
        if im != None:
          f.image = im
      me.faces.append(f)
      try:
        newmat = Material.Get('newmat')
      except:
        newmat = Material.New('newmat')
      #newmat.mode = newmat.mode  | 2048
      me.materials.append(newmat)
    #else:
    #  print "mesh ",triNode.name, " has no bitmap"
      
  return(me)

def eraseIPO (objectname):
        object = Blender.Object.Get(objectname)
        lIpo = object.getIpo()
        if lIpo != None:
                nbCurves = lIpo.getNcurves()
                for i in range(nbCurves):
                        nbBezPoints = lIpo.getNBezPoints(i)
                        for j in range(nbBezPoints):
                                lIpo.delBezPoint(i)


def GetOrCreateIPO(name):

        ipos = Blender.Ipo.Get()
        try:
                myipo = Blender.Ipo.Get(name)
        except:
                myipo = Blender.Ipo.New('Object',name)
        return myipo




def BuildMesh(mdlContent):
  rootdummy = None

  for nodename in mdlContent.model.nodes.keys():
    #print "  ", mdlContent.model.nodes[nodename].name
    curNode = mdlContent.model.nodes[nodename]
    #print curNode.type
    if curNode.type == 'trimeshnode':
      lamesh = BuildTrimeshNode(curNode)
      ob = NMesh.PutRaw(lamesh)
      ob.setName(curNode.name)
      #rot=Blender.Mathutils.Quaternion(curNode.orientation[0],curNode.orientation[1])

      #print curNode.name , curNode.orientation[1],4,"r", curNode.orientation[0]
      liste = [(curNode.orientation[0][0]), (curNode.orientation[0][1]),(curNode.orientation[0][2])]
      vecteur = Blender.Mathutils.Vector(liste)
      if  not (abs(curNode.orientation[1])<0.0001):
        rot = Blender.Mathutils.RotationMatrix(math.degrees(curNode.orientation[1]),4,"r", vecteur)
        ob.setMatrix(rot)
      dicObject[curNode.name] = ob.getName()

   
    elif curNode.type == 'dummynode':
      ob = CreateEmpty(curNode.name)
      if (rootdummy == None) and ((curNode.parent == 'NULL') or (curNode.parent == curNode.name)):
        rootdummy = ob
      dicObject[curNode.name] = ob.getName()

  for nodename in mdlContent.model.nodes.keys():
    if (mdlContent.model.nodes[nodename].parent != 'NULL') and (mdlContent.model.nodes[nodename].parent != mdlContent.model.nodes[nodename].name):
      father = Blender.Object.Get(dicObject[mdlContent.model.nodes[nodename].parent])
      father.makeParent([Blender.Object.Get(dicObject[mdlContent.model.nodes[nodename].name])],0,1)

  for nodename in mdlContent.model.nodes.keys():
    ob = Blender.Object.Get(dicObject[mdlContent.model.nodes[nodename].name])
    ob.setLocation (mdlContent.model.nodes[nodename].position)
    eraseIPO(ob.getName())
    
  #PROCESS ANIMS
  if len (mdlContent.anims) != 0:
    offsetframe = 1
    for anim in mdlContent.anims.keys():
      nokey = 0
      npkey = 0
      print "Anim : %s starts at frame %d" %(mdlContent.anims[anim].name, offsetframe)  
      for an in mdlContent.anims[anim].nodes.keys():
        if an in dicObject.keys():
          lipo = GetOrCreateIPO (an)
          Blender.Object.Get(dicObject[an]).setIpo(lipo)
        
          posX = GetOrCreateCurve(lipo,'LocX')
          posY = GetOrCreateCurve(lipo,'LocY')
          posZ = GetOrCreateCurve(lipo,'LocZ')
         
          #POSITION KEYS PROCESSING
          if len(mdlContent.anims[anim].nodes[an].positionsKeys) > 0:
            for pk in mdlContent.anims[anim].nodes[an].positionsKeys:
              insert_frame = offsetframe+pk[0]*25
              posX.addBezier ((insert_frame,pk[1][0]))
              posY.addBezier ((insert_frame,pk[1][1]))
              posZ.addBezier ((insert_frame,pk[1][2]))
          else:
            insert_frame = offsetframe
            posX.addBezier ((insert_frame,Blender.Object.Get(dicObject[an]).getLocation()[0]))
            posY.addBezier ((insert_frame,Blender.Object.Get(dicObject[an]).getLocation()[1]))
            posZ.addBezier ((insert_frame,Blender.Object.Get(dicObject[an]).getLocation()[2]))
          npkey = max (npkey,insert_frame)
          
          #ORIENTATION KEYS PROCESSING
          rotX = GetOrCreateCurve(lipo,'RotX')
          rotY = GetOrCreateCurve(lipo,'RotY')
          rotZ = GetOrCreateCurve(lipo,'RotZ')
            
          if len(mdlContent.anims[anim].nodes[an].orientationsKeys) > 0:
            for ok in mdlContent.anims[anim].nodes[an].orientationsKeys:
              insert_frame = offsetframe+ok[0]*25
              liste = [(ok[1][0]), (ok[1][1]),(ok[1][2])]
              vecteur = Blender.Mathutils.Vector(liste)
              try:
                rot = Blender.Mathutils.RotationMatrix(math.degrees(ok[2]),4,"r", vecteur)
              except:
                print "UNEXPECTED"
                rot = Blender.Mathutils.Matrix()
                rot.identity()
              eul = rot.rotationPart().toEuler()
              rotX.addBezier ((insert_frame,eul[0]/10))
              rotY.addBezier ((insert_frame,eul[1]/10))
              rotZ.addBezier ((insert_frame,eul[2]/10))
          else:
            insert_frame = offsetframe
            rotX.addBezier ((insert_frame,Blender.Object.Get(dicObject[an]).getLocation()[0]))
            rotY.addBezier ((insert_frame,Blender.Object.Get(dicObject[an]).getLocation()[1]))
            rotZ.addBezier ((insert_frame,Blender.Object.Get(dicObject[an]).getLocation()[2]))
          nokey  = max (nokey,insert_frame) 
          posX.setInterpolation('Linear')
          posX.setExtrapolation('Cyclic')
          posY.setInterpolation('Linear')
          posY.setExtrapolation('Cyclic')
          posZ.setInterpolation('Linear')
          posZ.setExtrapolation('Cyclic')
          
          posX.Recalc()
          posY.Recalc()
          posZ.Recalc()
          
          rotX.setInterpolation('Linear')
          rotX.setExtrapolation('Cyclic')
          rotY.setInterpolation('Linear')
          rotY.setExtrapolation('Cyclic')
          rotZ.setInterpolation('Linear')
          rotZ.setExtrapolation('Cyclic')
            
          rotX.Recalc()
          rotY.Recalc()
          rotZ.Recalc()
      offsetframe = max(nokey, npkey)
  

  return rootdummy

#**************************
#**** MAIN PROGRAM ********
#**************************

def LoadMDL (filename):
  global filepath, thefilename
  thefilename = split(sys.basename(filename),'.')[0]
  print "Loading file", filename
  filepath = sys.dirname (filename)
  ficin = open (filename)
  lignes = ficin.readlines()
  for ligne in lignes:
    ProcessLine(ligne)
  ficin.close()
  return (BuildMesh (currentContent))

def LoadAndPlaceMDLFile(pfilename,posX, posY, posZ, Orientation):
  themesh= LoadMDL(pfilename)
  
  themesh.setLocation(posX*10, posY*10, posZ*5)
  if Orientation == 0:
    Angle = 0
  elif Orientation == 1:
    Angle = math.radians(90)
  elif Orientation == 2:
    Angle = math.radians(180)
  elif Orientation == 3:
    Angle = math.radians(270)
  themesh.setEuler(Blender.Mathutils.Euler([0.0, 0.0, Angle]))
  Blender.Redraw()

def LoadAre (filename):
  CheminProjet = 'c:\\Documents and Settings\\Jean-Baptiste\\Mes Documents\\Mes Projets\\Aurora2Blend'
  fic = open (filename,'r')
  lines = fic.readlines ()
  fic.close()
  for l in lines:
    params = l.split()
    #print CheminProjet+'\\Modeles\\'+params[0],params[1],params[2],params[3],params[4] 
    LoadAndPlaceMDLFile(CheminProjet+'\\Modeles\\'+params[0],int(params[1]),int(params[2]),int(params[3]),int(params[4]))
  Blender.Redraw()


def ProcessFile (name):
  extension =  splitext (basename(name))[1]
  if extension == ".txt":
    LoadAre (name)
  elif extension == ".mdl":
    LoadAndPlaceMDLFile(name,0,0,0,0)


Window.FileSelector(ProcessFile, "Neverwinter MDL file")
